// This file is licensed under the Elastic License 2.0. Copyright 2021 StarRocks Limited.

#include "exprs/vectorized/array_element_expr.h"

#include "column/array_column.h"
#include "column/column_helper.h"
#include "column/fixed_length_column.h"
#include "common/object_pool.h"
#include "util/raw_container.h"

namespace starrocks::vectorized {

class ArrayElementExpr final : public Expr {
public:
    explicit ArrayElementExpr(const TExprNode& node) : Expr(node) {}

    ArrayElementExpr(const ArrayElementExpr&) = default;
    ArrayElementExpr(ArrayElementExpr&&) = default;

    ColumnPtr evaluate(ExprContext* context, vectorized::Chunk* chunk) override {
        DCHECK_EQ(2, _children.size());
        DCHECK_EQ(_type, _children[0]->type().children[0]);
        ColumnPtr arg0 = _children[0]->evaluate(context, chunk);
        ColumnPtr arg1 = _children[1]->evaluate(context, chunk);
        size_t num_rows = std::max(arg0->size(), arg1->size());
        // No optimization for const column now.
        arg0 = ColumnHelper::unfold_const_column(_children[0]->type(), num_rows, arg0);
        arg1 = ColumnHelper::unfold_const_column(_children[1]->type(), num_rows, arg1);
        auto* array_column = down_cast<ArrayColumn*>(get_data_column(arg0.get()));
        auto* array_elements = array_column->elements_column().get();
        auto* array_elements_data = get_data_column(array_elements);
        DCHECK_EQ(num_rows, arg0->size());
        DCHECK_EQ(num_rows, arg1->size());
        DCHECK_EQ(num_rows + 1, array_column->offsets_column()->size());

        const int32_t* subscripts = down_cast<Int32Column*>(get_data_column(arg1.get()))->get_data().data();
        const uint32_t* offsets = array_column->offsets_column()->get_data().data();

        std::vector<uint8_t> null_flags;
        raw::make_room(&null_flags, num_rows);

        // Construct null flags.
        uint32_t prev = offsets[0];
        for (size_t i = 1; i <= num_rows; i++) {
            uint32_t curr = offsets[i];
            DCHECK_GE(curr, prev);
            // cast negative subscript to large integer.
            uint32_t subscript = (uint32_t)subscripts[i - 1];
            null_flags[i - 1] = (subscript > (curr - prev)) | (subscript == 0);
            prev = curr;
        }

        if (auto* nullable = dynamic_cast<NullableColumn*>(arg0.get()); nullable != nullptr) {
            const uint8_t* nulls = nullable->null_column()->raw_data();
            for (size_t i = 0; i < num_rows; i++) {
                null_flags[i] |= nulls[i];
            }
        }
        if (auto* nullable = dynamic_cast<NullableColumn*>(arg1.get()); nullable != nullptr) {
            const uint8_t* nulls = nullable->null_column()->raw_data();
            for (size_t i = 0; i < num_rows; i++) {
                null_flags[i] |= nulls[i];
            }
        }

        // construct selection list.
        std::vector<uint32_t> selection;
        starrocks::raw::make_room(&selection, num_rows);

        prev = offsets[0];
        uint32_t idx = 0;
        for (size_t i = 1; i <= num_rows; i++) {
            uint32_t curr = offsets[i];
            idx = null_flags[i - 1] ? idx : prev + (subscripts[i - 1] - 1);
            selection[i - 1] = idx;
            prev = curr;
        }
        DCHECK_EQ(num_rows, selection.size());

        if (array_elements->has_null()) {
            auto* nullable_elements = down_cast<NullableColumn*>(array_elements);
            const uint8_t* nulls = nullable_elements->null_column()->raw_data();
            for (size_t i = 0; i < num_rows; i++) {
                null_flags[i] |= nulls[selection[i]];
            }
        }

        // Construct the final result column;
        ColumnPtr result_data = array_elements_data->clone_empty();
        NullColumnPtr result_null = NullColumn::create();
        result_null->get_data().swap(null_flags);

        if (!array_elements_data->empty()) {
            result_data->append_selective(*array_elements_data, selection.data(), 0, num_rows);
        } else {
            result_data->append_default(num_rows);
        }
        DCHECK_EQ(result_null->size(), result_data->size());

        return NullableColumn::create(std::move(result_data), std::move(result_null));
    }

    Expr* clone(ObjectPool* pool) const override { return pool->add(new ArrayElementExpr(*this)); }

private:
    Column* get_data_column(Column* column) { return ColumnHelper::get_data_column(column); }
};

Expr* ArrayElementExprFactory::from_thrift(const TExprNode& node) {
    DCHECK_EQ(TExprNodeType::ARRAY_ELEMENT_EXPR, node.node_type);
    return new ArrayElementExpr(node);
}

} // namespace starrocks::vectorized
