// This file is licensed under the Elastic License 2.0. Copyright 2021 StarRocks Limited.

#include "chunk_cursor.h"

#include "column/chunk.h"
#include "exec/sort_exec_exprs.h"

namespace starrocks::vectorized {

ChunkCursor::ChunkCursor(const ChunkSupplier& chunk_supplier, const ChunkProbeSupplier& chunk_probe_supplier,
                         const ChunkHasSupplier& chunk_has_supplier, const std::vector<ExprContext*>* exprs,
                         const std::vector<bool>* is_asc, const std::vector<bool>* is_null_first, bool is_pipeline)
        : _chunk_supplier(chunk_supplier),
          _chunk_probe_supplier(chunk_probe_supplier),
          _chunk_has_supplier(chunk_has_supplier),
          _current_pos(-1),
          _sort_exprs(exprs),
          _is_pipeline(is_pipeline) {
    DCHECK_EQ(_sort_exprs->size(), is_asc->size());
    DCHECK_EQ(is_asc->size(), is_null_first->size());

    size_t col_num = is_asc->size();
    _sort_order_flag.resize(col_num);
    _null_first_flag.resize(col_num);
    for (size_t i = 0; i < col_num; ++i) {
        _sort_order_flag[i] = (*is_asc)[i] ? 1 : -1;
        if ((*is_asc)[i]) {
            _null_first_flag[i] = (*is_null_first)[i] ? -1 : 1;
        } else {
            _null_first_flag[i] = (*is_null_first)[i] ? 1 : -1;
        }
    }

    if (!_is_pipeline) {
        _reset_with_next_chunk();
    }
}

ChunkCursor::~ChunkCursor() {}

bool ChunkCursor::operator<(const ChunkCursor& cursor) const {
    DCHECK_EQ(_current_order_by_columns.size(), cursor._current_order_by_columns.size());
    // both cursors must be pointing to valid data.
    DCHECK(_current_pos >= 0 && _current_chunk != nullptr);
    DCHECK(cursor._current_pos >= 0 && cursor._current_chunk != nullptr);
    const size_t number_of_order_by_columns = _current_order_by_columns.size();
    bool is_ahead = true;
    for (size_t col_index = 0; col_index < number_of_order_by_columns; ++col_index) {
        const auto& left_col = _current_order_by_columns[col_index];
        const auto& right_col = cursor._current_order_by_columns[col_index];
        int cmp = left_col->compare_at(_current_pos, cursor._current_pos, *right_col, _null_first_flag[col_index]);
        if (cmp != 0) {
            if (_sort_order_flag[col_index] > 0) {
                is_ahead = cmp < 0;
            } else {
                is_ahead = cmp > 0;
            }
            break;
        }
    }
    return is_ahead;
}

bool ChunkCursor::is_valid() const {
    return _current_pos >= 0 && _current_chunk != nullptr;
}

void ChunkCursor::next() {
    if (_current_chunk == nullptr) {
        return;
    }
    ++_current_pos;
    if (_current_pos >= _current_chunk->num_rows()) {
        _reset_with_next_chunk();
        if (_current_chunk != nullptr) {
            ++_current_pos;
        }
    }
}

bool ChunkCursor::has_next() {
    if (_current_chunk == nullptr) {
        return false;
    }

    if ((_current_pos + 1) < _current_chunk->num_rows()) {
        return true;
    }

    return false;
}

void ChunkCursor::next_for_pipeline() {
    if (_current_chunk == nullptr) {
        return;
    }
    ++_current_pos;
    if (_current_pos >= _current_chunk->num_rows()) {
        reset_with_next_chunk_for_pipeline();
        if (_current_chunk != nullptr) {
            ++_current_pos;
        }
    }
}

ChunkPtr ChunkCursor::clone_empty_chunk(size_t reserved_row_number) const {
    if (_current_chunk == nullptr) {
        return nullptr;
    } else {
        return _current_chunk->clone_empty_with_slot(reserved_row_number);
    }
}

bool ChunkCursor::copy_current_row_to(Chunk* dest) const {
    dest->append(*_current_chunk, _current_pos, 1);
    return true;
}

Status ChunkCursor::chunk_supplier(Chunk** chunk) {
    return _chunk_supplier(chunk);
}

bool ChunkCursor::chunk_probe_supplier(Chunk** chunk) {
    return _chunk_probe_supplier(chunk);
}

bool ChunkCursor::chunk_has_supplier() {
    return _chunk_has_supplier();
}

void ChunkCursor::_reset_with_next_chunk() {
    _current_order_by_columns.clear();
    Chunk* tmp_chunk = nullptr;
    _chunk_supplier(&tmp_chunk);
    _current_chunk.reset(tmp_chunk);
    _current_pos = -1;
    if (_current_chunk == nullptr) {
        return;
    }

    // prepare order by columns
    _current_order_by_columns.reserve(_sort_exprs->size());
    for (ExprContext* expr_ctx : *_sort_exprs) {
        _current_order_by_columns.push_back(expr_ctx->evaluate(_current_chunk.get()));
    }
}

void ChunkCursor::reset_with_next_chunk_for_pipeline() {
    _current_order_by_columns.clear();
    Chunk* tmp_chunk = nullptr;
    _chunk_probe_supplier(&tmp_chunk);
    _current_chunk.reset(tmp_chunk);
    _current_pos = -1;
    if (_current_chunk == nullptr) {
        return;
    }
    // prepare order by columns
    _current_order_by_columns.reserve(_sort_exprs->size());
    for (ExprContext* expr_ctx : *_sort_exprs) {
        _current_order_by_columns.push_back(expr_ctx->evaluate(_current_chunk.get()));
    }
}

} // namespace starrocks::vectorized
