// This file is made available under Elastic License 2.0.
// This file is based on code available under the Apache license here:
//   https://github.com/apache/incubator-doris/blob/master/fe/fe-core/src/test/java/org/apache/doris/load/loadv2/LoadingTaskPlannerTest.java

// Licensed to the Apache Software Foundation (ASF) under one
// or more contributor license agreements.  See the NOTICE file
// distributed with this work for additional information
// regarding copyright ownership.  The ASF licenses this file
// to you under the Apache License, Version 2.0 (the
// "License"); you may not use this file except in compliance
// with the License.  You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

package com.starrocks.load.loadv2;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.starrocks.analysis.BrokerDesc;
import com.starrocks.analysis.DataDescription;
import com.starrocks.analysis.Expr;
import com.starrocks.analysis.FunctionName;
import com.starrocks.analysis.LoadStmt;
import com.starrocks.analysis.SqlParser;
import com.starrocks.analysis.SqlScanner;
import com.starrocks.catalog.AggregateType;
import com.starrocks.catalog.Catalog;
import com.starrocks.catalog.Column;
import com.starrocks.catalog.Database;
import com.starrocks.catalog.Function;
import com.starrocks.catalog.KeysType;
import com.starrocks.catalog.OlapTable;
import com.starrocks.catalog.Partition;
import com.starrocks.catalog.ScalarType;
import com.starrocks.catalog.Type;
import com.starrocks.common.Config;
import com.starrocks.common.UserException;
import com.starrocks.common.jmockit.Deencapsulation;
import com.starrocks.common.util.SqlParserUtils;
import com.starrocks.common.util.TimeUtils;
import com.starrocks.load.BrokerFileGroup;
import com.starrocks.load.Load;
import com.starrocks.planner.FileScanNode;
import com.starrocks.planner.OlapTableSink;
import com.starrocks.planner.PlanFragment;
import com.starrocks.system.Backend;
import com.starrocks.system.SystemInfoService;
import com.starrocks.thrift.TBrokerFileStatus;
import com.starrocks.thrift.TBrokerScanRangeParams;
import com.starrocks.thrift.TExpr;
import com.starrocks.thrift.TExprNode;
import com.starrocks.thrift.TExprNodeType;
import com.starrocks.thrift.TOpType;
import com.starrocks.thrift.TPlanFragment;
import com.starrocks.thrift.TPlanNode;
import com.starrocks.thrift.TPlanNodeType;
import com.starrocks.thrift.TPrimitiveType;
import com.starrocks.thrift.TScanRangeLocations;
import com.starrocks.thrift.TUniqueId;
import mockit.Expectations;
import mockit.Injectable;
import mockit.Mocked;
import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

import java.io.StringReader;
import java.util.Arrays;
import java.util.List;
import java.util.Map;

public class LoadingTaskPlannerTest {
    private long jobId;
    private long txnId;
    private TUniqueId loadId;
    private BrokerDesc brokerDesc;

    // config
    private int loadParallelInstanceNum;
    private int maxBrokerConcurrency;
    private boolean enableVectorizedLoad;
    private boolean enableVectorizedFileLoad;

    // backends
    private ImmutableMap<Long, Backend> idToBackend;

    @Mocked
    Partition partition;
    @Mocked
    OlapTableSink sink;

    @Before
    public void setUp() {
        jobId = 1L;
        txnId = 2L;
        loadId = new TUniqueId(3, 4);
        brokerDesc = new BrokerDesc("broker0", null);

        loadParallelInstanceNum = Config.load_parallel_instance_num;
        maxBrokerConcurrency = Config.max_broker_concurrency;
        enableVectorizedLoad = Config.vectorized_load_enable;
        enableVectorizedFileLoad = Config.enable_vectorized_file_load;

        // backends
        Map<Long, Backend> idToBackendTmp = Maps.newHashMap();
        Backend b1 = new Backend(0L, "host0", 9050);
        b1.setAlive(true);
        idToBackendTmp.put(0L, b1);
        Backend b2 = new Backend(1L, "host1", 9050);
        b2.setAlive(true);
        idToBackendTmp.put(1L, b2);
        idToBackend = ImmutableMap.copyOf(idToBackendTmp);
    }

    @After
    public void tearDown() {
        Config.load_parallel_instance_num = loadParallelInstanceNum;
        Config.max_broker_concurrency = maxBrokerConcurrency;
        Config.vectorized_load_enable = enableVectorizedLoad;
        Config.enable_vectorized_file_load = enableVectorizedFileLoad;
    }

    @Test
    public void testParallelInstance(@Mocked Catalog catalog, @Mocked SystemInfoService systemInfoService,
                                     @Injectable Database db, @Injectable OlapTable table) throws UserException {
        // table schema
        List<Column> columns = Lists.newArrayList();
        Column c1 = new Column("c1", Type.BIGINT, true);
        columns.add(c1);
        Column c2 = new Column("c2", Type.BIGINT, true);
        columns.add(c2);
        List<String> columnNames = Lists.newArrayList("c1", "c2");

        new Expectations() {
            {
                Catalog.getCurrentSystemInfo();
                result = systemInfoService;
                systemInfoService.getIdToBackend();
                result = idToBackend;
                table.getBaseSchema();
                result = columns;
                table.getFullSchema();
                result = columns;
                table.getPartitions();
                minTimes = 0;
                result = Arrays.asList(partition);
                partition.getId();
                minTimes = 0;
                result = 0;
                table.getColumn("c1");
                result = columns.get(0);
                table.getColumn("c2");
                result = columns.get(1);
            }
        };

        // file groups
        List<BrokerFileGroup> fileGroups = Lists.newArrayList();
        List<String> files = Lists.newArrayList("hdfs://127.0.0.1:9001/file1", "hdfs://127.0.0.1:9001/file2");
        DataDescription desc =
                new DataDescription("testTable", null, files, columnNames, null, null, null, false, null);
        BrokerFileGroup brokerFileGroup = new BrokerFileGroup(desc);
        Deencapsulation.setField(brokerFileGroup, "columnSeparator", "\t");
        Deencapsulation.setField(brokerFileGroup, "rowDelimiter", "\n");
        fileGroups.add(brokerFileGroup);

        // file status
        List<List<TBrokerFileStatus>> fileStatusesList = Lists.newArrayList();
        List<TBrokerFileStatus> fileStatusList = Lists.newArrayList();
        fileStatusList.add(new TBrokerFileStatus("hdfs://127.0.0.1:9001/file1", false, 268435456, true));
        fileStatusList.add(new TBrokerFileStatus("hdfs://127.0.0.1:9001/file2", false, 268435456, true));
        fileStatusesList.add(fileStatusList);

        // load_parallel_instance_num: 1
        Config.load_parallel_instance_num = 1;
        Config.vectorized_load_enable = true;
        Config.enable_vectorized_file_load = true;
        LoadingTaskPlanner planner = new LoadingTaskPlanner(jobId, txnId, db.getId(), table, brokerDesc, fileGroups,
                false, TimeUtils.DEFAULT_TIME_ZONE, 3600);
        planner.plan(loadId, fileStatusesList, 2);
        Assert.assertEquals(1, planner.getScanNodes().size());
        FileScanNode scanNode = (FileScanNode) planner.getScanNodes().get(0);
        List<TScanRangeLocations> locationsList = scanNode.getScanRangeLocations(0);
        Assert.assertEquals(2, locationsList.size());

        // load_parallel_instance_num: 2
        Config.load_parallel_instance_num = 2;
        planner = new LoadingTaskPlanner(jobId, txnId, db.getId(), table, brokerDesc, fileGroups,
                false, TimeUtils.DEFAULT_TIME_ZONE, 3600);
        planner.plan(loadId, fileStatusesList, 2);
        scanNode = (FileScanNode) planner.getScanNodes().get(0);
        locationsList = scanNode.getScanRangeLocations(0);
        Assert.assertEquals(4, locationsList.size());

        // load_parallel_instance_num: 2, max_broker_concurrency: 3
        Config.load_parallel_instance_num = 2;
        Config.max_broker_concurrency = 3;
        planner = new LoadingTaskPlanner(jobId, txnId, db.getId(), table, brokerDesc, fileGroups,
                false, TimeUtils.DEFAULT_TIME_ZONE, 3600);
        planner.plan(loadId, fileStatusesList, 2);
        scanNode = (FileScanNode) planner.getScanNodes().get(0);
        locationsList = scanNode.getScanRangeLocations(0);
        Assert.assertEquals(3, locationsList.size());
    }

    @Test
    public void testVectorizedLoad(@Mocked Catalog catalog, @Mocked SystemInfoService systemInfoService,
                                   @Injectable Database db, @Injectable OlapTable table) throws Exception {
        // table schema
        List<Column> columns = Lists.newArrayList();
        columns.add(new Column("k1", Type.TINYINT, true, null, true, null, ""));
        columns.add(new Column("k2", Type.INT, true, null, false, null, ""));
        columns.add(new Column("k3", ScalarType.createVarchar(50), true, null, true, null, ""));
        columns.add(new Column("v", Type.BIGINT, false, AggregateType.SUM, false, null, ""));

        Function f1 = new Function(new FunctionName("substr"), new Type[] {Type.VARCHAR, Type.INT, Type.INT},
                Type.VARCHAR, true);
        f1.setIsVectorized(true);
        Function f2 = new Function(new FunctionName("casttoint"), new Type[] {Type.VARCHAR},
                Type.INT, true);
        f2.setIsVectorized(true);
        new Expectations() {
            {
                Catalog.getCurrentSystemInfo();
                result = systemInfoService;
                systemInfoService.getIdToBackend();
                result = idToBackend;
                table.getBaseSchema();
                result = columns;
                table.getFullSchema();
                result = columns;
                table.getPartitions();
                minTimes = 0;
                result = Arrays.asList(partition);
                partition.getId();
                minTimes = 0;
                result = 0;
                table.getColumn("k1");
                result = columns.get(0);
                table.getColumn("k2");
                result = columns.get(1);
                table.getColumn("k3");
                result = columns.get(2);
                table.getColumn("v");
                result = columns.get(3);
                table.getColumn("k33");
                result = null;
                catalog.getFunction((Function) any, (Function.CompareMode) any);
                returns(f1, f1, f2);
            }
        };

        // column mappings
        String sql = "LOAD LABEL label0 (DATA INFILE('path/k2=1/file1') INTO TABLE t2 FORMAT AS 'orc' (k1,k33,v) " +
                "COLUMNS FROM PATH AS (k2) set (k3 = substr(k33,1,5))) WITH BROKER 'broker0'";
        SqlParser parser = new SqlParser(new SqlScanner(new StringReader(sql)));
        LoadStmt loadStmt = (LoadStmt) SqlParserUtils.getFirstStmt(parser);
        List<Expr> columnMappingList = Deencapsulation.getField(loadStmt.getDataDescriptions().get(0),
                "columnMappingList");

        // file groups
        List<BrokerFileGroup> fileGroups = Lists.newArrayList();
        List<String> files = Lists.newArrayList("path/k2=1/file1");
        List<String> columnNames = Lists.newArrayList("k1", "k33", "v");
        DataDescription desc = new DataDescription("t2", null, files, columnNames,
                null, null, "ORC", Lists.newArrayList("k2"),
                false, columnMappingList, null);
        Deencapsulation.invoke(desc, "analyzeColumns");
        BrokerFileGroup brokerFileGroup = new BrokerFileGroup(desc);
        Deencapsulation.setField(brokerFileGroup, "columnSeparator", "\t");
        Deencapsulation.setField(brokerFileGroup, "rowDelimiter", "\n");
        Deencapsulation.setField(brokerFileGroup, "fileFormat", "ORC");
        fileGroups.add(brokerFileGroup);

        // file status
        List<List<TBrokerFileStatus>> fileStatusesList = Lists.newArrayList();
        List<TBrokerFileStatus> fileStatusList = Lists.newArrayList();
        fileStatusList.add(new TBrokerFileStatus("path/k2=1/file1", false, 268435456, true));
        fileStatusesList.add(fileStatusList);

        // plan
        Config.vectorized_load_enable = true;
        Config.enable_vectorized_file_load = true;
        LoadingTaskPlanner planner = new LoadingTaskPlanner(jobId, txnId, db.getId(), table, brokerDesc, fileGroups,
                false, TimeUtils.DEFAULT_TIME_ZONE, 3600);
        planner.plan(loadId, fileStatusesList, 1);

        // 1. check fragment
        List<PlanFragment> fragments = planner.getFragments();
        Assert.assertEquals(1, fragments.size());
        PlanFragment fragment = fragments.get(0);
        TPlanFragment tPlanFragment = fragment.toThrift();
        List<TPlanNode> nodes = tPlanFragment.plan.nodes;
        Assert.assertEquals(1, nodes.size());
        TPlanNode tPlanNode = nodes.get(0);
        Assert.assertEquals(TPlanNodeType.FILE_SCAN_NODE, tPlanNode.node_type);
        Assert.assertTrue(tPlanNode.use_vectorized);

        // 2. check scan node column expr
        FileScanNode scanNode = (FileScanNode) planner.getScanNodes().get(0);
        List<TScanRangeLocations> locationsList = scanNode.getScanRangeLocations(0);
        Assert.assertEquals(1, locationsList.size());
        TScanRangeLocations location = locationsList.get(0);
        TBrokerScanRangeParams params = location.scan_range.broker_scan_range.params;
        Map<Integer, TExpr> exprOfDestSlot = params.expr_of_dest_slot;

        // 2.1 check k1
        TExpr k1Expr = exprOfDestSlot.get(0);
        Assert.assertEquals(1, k1Expr.nodes.size());
        TExprNode node = k1Expr.nodes.get(0);
        Assert.assertEquals(TExprNodeType.SLOT_REF, node.node_type);
        Assert.assertEquals(TPrimitiveType.TINYINT, node.type.types.get(0).scalar_type.type);
        Assert.assertTrue(node.use_vectorized);

        // 2.2 check k2 from path
        TExpr k2Expr = exprOfDestSlot.get(1);
        Assert.assertEquals(2, k2Expr.nodes.size());
        TExprNode castNode = k2Expr.nodes.get(0);
        Assert.assertEquals(TExprNodeType.CAST_EXPR, castNode.node_type);
        Assert.assertEquals(TPrimitiveType.INT, castNode.fn.ret_type.types.get(0).scalar_type.type);
        Assert.assertTrue(castNode.use_vectorized);
        node = k2Expr.nodes.get(1);
        Assert.assertEquals(TExprNodeType.SLOT_REF, node.node_type);
        Assert.assertEquals(TPrimitiveType.VARCHAR, node.type.types.get(0).scalar_type.type);
        Assert.assertTrue(node.use_vectorized);

        // 2.3 check k3 mapping
        TExpr k3Expr = exprOfDestSlot.get(2);
        Assert.assertEquals(4, k3Expr.nodes.size());
        node = k3Expr.nodes.get(0);
        Assert.assertEquals(TExprNodeType.FUNCTION_CALL, node.node_type);
        Assert.assertEquals("substr", node.fn.name.function_name);
        node = k3Expr.nodes.get(1);
        Assert.assertEquals(TExprNodeType.SLOT_REF, node.node_type);
        Assert.assertEquals(TPrimitiveType.VARCHAR, node.type.types.get(0).scalar_type.type);
        node = k3Expr.nodes.get(2);
        Assert.assertEquals(TExprNodeType.INT_LITERAL, node.node_type);
        Assert.assertEquals(1, node.int_literal.value);
        node = k3Expr.nodes.get(3);
        Assert.assertEquals(TExprNodeType.INT_LITERAL, node.node_type);
        Assert.assertEquals(5, node.int_literal.value);
        Assert.assertTrue(node.use_vectorized);
    }

    @Test
    public void testLoadWithOpColumnDefault(@Mocked Catalog catalog, @Mocked SystemInfoService systemInfoService,
                                            @Injectable Database db, @Injectable OlapTable table) throws Exception {
        // table schema
        List<Column> columns = Lists.newArrayList();
        columns.add(new Column("pk", Type.BIGINT, true, null, false, null, ""));
        columns.add(new Column("v1", Type.INT, false, null, false, null, ""));
        columns.add(new Column("v2", ScalarType.createVarchar(50), false, null, true, null, ""));

        new Expectations() {
            {
                Catalog.getCurrentSystemInfo();
                result = systemInfoService;
                systemInfoService.getIdToBackend();
                result = idToBackend;
                table.getKeysType();
                result = KeysType.PRIMARY_KEYS;
                table.getBaseSchema();
                result = columns;
                table.getFullSchema();
                result = columns;
                table.getPartitions();
                minTimes = 0;
                result = Arrays.asList(partition);
                partition.getId();
                minTimes = 0;
                result = 0;
                table.getColumn("pk");
                result = columns.get(0);
                table.getColumn("v1");
                result = columns.get(1);
                table.getColumn("v2");
                result = columns.get(2);
                table.getColumn(Load.LOAD_OP_COLUMN);
                result = null;
            }
        };

        // column mappings
        String sql =
                "LOAD LABEL label0 (DATA INFILE('/path/file1') INTO TABLE t2 columns terminated by ',') with broker 'broker0'";
        SqlParser parser = new SqlParser(new SqlScanner(new StringReader(sql)));
        LoadStmt loadStmt = (LoadStmt) SqlParserUtils.getFirstStmt(parser);
        List<Expr> columnMappingList = Deencapsulation.getField(loadStmt.getDataDescriptions().get(0),
                "columnMappingList");

        // file groups
        List<BrokerFileGroup> fileGroups = Lists.newArrayList();
        List<String> files = Lists.newArrayList("/path/file1");
        List<String> columnNames = Lists.newArrayList("pk", "v1", "v2");
        DataDescription desc = new DataDescription("t2", null, files, columnNames,
                null, null, "CSV", Lists.newArrayList(),
                false, columnMappingList, null);
        Deencapsulation.invoke(desc, "analyzeColumns");
        BrokerFileGroup brokerFileGroup = new BrokerFileGroup(desc);
        Deencapsulation.setField(brokerFileGroup, "columnSeparator", ",");
        Deencapsulation.setField(brokerFileGroup, "rowDelimiter", "\n");
        Deencapsulation.setField(brokerFileGroup, "fileFormat", "CSV");
        fileGroups.add(brokerFileGroup);

        // file status
        List<List<TBrokerFileStatus>> fileStatusesList = Lists.newArrayList();
        List<TBrokerFileStatus> fileStatusList = Lists.newArrayList();
        fileStatusList.add(new TBrokerFileStatus("/path/file1", false, 128000000, true));
        fileStatusesList.add(fileStatusList);

        // plan
        Config.vectorized_load_enable = true;
        LoadingTaskPlanner planner = new LoadingTaskPlanner(jobId, txnId, db.getId(), table, brokerDesc, fileGroups,
                false, TimeUtils.DEFAULT_TIME_ZONE, 3600);
        planner.plan(loadId, fileStatusesList, 1);

        // 2. check scan node column expr
        FileScanNode scanNode = (FileScanNode) planner.getScanNodes().get(0);
        List<TScanRangeLocations> locationsList = scanNode.getScanRangeLocations(0);
        Assert.assertEquals(1, locationsList.size());
        TScanRangeLocations location = locationsList.get(0);
        TBrokerScanRangeParams params = location.scan_range.broker_scan_range.params;
        Map<Integer, TExpr> exprOfDestSlot = params.expr_of_dest_slot;

        // get last slot: id == 3
        TExpr opExpr = exprOfDestSlot.get(3);
        Assert.assertEquals(1, opExpr.nodes.size());
        Assert.assertEquals(TExprNodeType.INT_LITERAL, opExpr.nodes.get(0).node_type);
    }

    @Test
    public void testLoadWithOpColumnDelete(@Mocked Catalog catalog, @Mocked SystemInfoService systemInfoService,
                                           @Injectable Database db, @Injectable OlapTable table) throws Exception {
        // table schema
        List<Column> columns = Lists.newArrayList();
        columns.add(new Column("pk", Type.BIGINT, true, null, false, null, ""));
        columns.add(new Column("v1", Type.INT, false, null, false, null, ""));
        columns.add(new Column("v2", ScalarType.createVarchar(50), false, null, true, null, ""));

        new Expectations() {
            {
                Catalog.getCurrentSystemInfo();
                result = systemInfoService;
                systemInfoService.getIdToBackend();
                result = idToBackend;
                table.getKeysType();
                result = KeysType.PRIMARY_KEYS;
                table.getBaseSchema();
                result = columns;
                table.getFullSchema();
                result = columns;
                table.getPartitions();
                minTimes = 0;
                result = Arrays.asList(partition);
                partition.getId();
                minTimes = 0;
                result = 0;
                table.getColumn("pk");
                result = columns.get(0);
                table.getColumn("v1");
                result = columns.get(1);
                table.getColumn("v2");
                result = columns.get(2);
                table.getColumn(Load.LOAD_OP_COLUMN);
                result = null;
            }
        };

        // column mappings
        String sql =
                "LOAD LABEL label0 (DATA INFILE('/path/file1') INTO TABLE t2 columns terminated by ',' set ( __op = 'delete')) with broker 'broker0'";
        SqlParser parser = new SqlParser(new SqlScanner(new StringReader(sql)));
        LoadStmt loadStmt = (LoadStmt) SqlParserUtils.getFirstStmt(parser);
        List<Expr> columnMappingList = Deencapsulation.getField(loadStmt.getDataDescriptions().get(0),
                "columnMappingList");

        // file groups
        List<BrokerFileGroup> fileGroups = Lists.newArrayList();
        List<String> files = Lists.newArrayList("/path/file1");
        List<String> columnNames = Lists.newArrayList("pk", "v1", "v2");
        DataDescription desc = new DataDescription("t2", null, files, columnNames,
                null, null, "CSV", Lists.newArrayList(),
                false, columnMappingList, null);
        Deencapsulation.invoke(desc, "analyzeColumns");
        BrokerFileGroup brokerFileGroup = new BrokerFileGroup(desc);
        Deencapsulation.setField(brokerFileGroup, "columnSeparator", ",");
        Deencapsulation.setField(brokerFileGroup, "rowDelimiter", "\n");
        Deencapsulation.setField(brokerFileGroup, "fileFormat", "CSV");
        fileGroups.add(brokerFileGroup);

        // file status
        List<List<TBrokerFileStatus>> fileStatusesList = Lists.newArrayList();
        List<TBrokerFileStatus> fileStatusList = Lists.newArrayList();
        fileStatusList.add(new TBrokerFileStatus("/path/file1", false, 128000000, true));
        fileStatusesList.add(fileStatusList);

        // plan
        Config.vectorized_load_enable = true;
        LoadingTaskPlanner planner = new LoadingTaskPlanner(jobId, txnId, db.getId(), table, brokerDesc, fileGroups,
                false, TimeUtils.DEFAULT_TIME_ZONE, 3600);
        planner.plan(loadId, fileStatusesList, 1);

        // 2. check scan node column expr
        FileScanNode scanNode = (FileScanNode) planner.getScanNodes().get(0);
        List<TScanRangeLocations> locationsList = scanNode.getScanRangeLocations(0);
        Assert.assertEquals(1, locationsList.size());
        TScanRangeLocations location = locationsList.get(0);
        TBrokerScanRangeParams params = location.scan_range.broker_scan_range.params;
        Map<Integer, TExpr> exprOfDestSlot = params.expr_of_dest_slot;

        // get last slot: id == 3
        TExpr opExpr = exprOfDestSlot.get(3);
        Assert.assertEquals(1, opExpr.nodes.size());
        Assert.assertEquals(TExprNodeType.INT_LITERAL, opExpr.nodes.get(0).node_type);
        Assert.assertEquals(TOpType.DELETE.getValue(), opExpr.nodes.get(0).int_literal.value);
    }

    @Test
    public void testLoadWithOpColumnExpr(@Mocked Catalog catalog, @Mocked SystemInfoService systemInfoService,
                                         @Injectable Database db, @Injectable OlapTable table) throws Exception {
        // table schema
        List<Column> columns = Lists.newArrayList();
        columns.add(new Column("pk", Type.BIGINT, true, null, false, null, ""));
        columns.add(new Column("v1", Type.INT, false, null, false, null, ""));
        columns.add(new Column("v2", ScalarType.createVarchar(50), false, null, true, null, ""));

        Function f1 = new Function(new FunctionName("casttobigint"), new Type[] {Type.VARCHAR},
                Type.BIGINT, true);
        f1.setIsVectorized(true);
        Function f2 = new Function(new FunctionName("casttoint"), new Type[] {Type.VARCHAR},
                Type.INT, true);
        f2.setIsVectorized(true);
        Function f3 = new Function(new FunctionName("casttotinyint"), new Type[] {Type.VARCHAR},
                Type.TINYINT, true);
        f3.setIsVectorized(true);

        new Expectations() {
            {
                Catalog.getCurrentSystemInfo();
                result = systemInfoService;
                systemInfoService.getIdToBackend();
                result = idToBackend;
                table.getKeysType();
                result = KeysType.PRIMARY_KEYS;
                table.getBaseSchema();
                result = columns;
                table.getFullSchema();
                result = columns;
                table.getPartitions();
                minTimes = 0;
                result = Arrays.asList(partition);
                partition.getId();
                minTimes = 0;
                result = 0;
                table.getColumn("c0");
                result = null;
                table.getColumn("c1");
                result = null;
                table.getColumn("c2");
                result = null;
                table.getColumn("c3");
                result = null;
                result = columns.get(0);
                table.getColumn("pk");
                result = columns.get(0);
                table.getColumn("v1");
                result = columns.get(1);
                table.getColumn("v2");
                result = columns.get(2);
                table.getColumn(Load.LOAD_OP_COLUMN);
                result = null;
                catalog.getFunction((Function) any, (Function.CompareMode) any);
                returns(f1, f2, f3);
            }
        };

        // column mappings
        String sql =
                "LOAD LABEL label0 (DATA INFILE('/path/file1') INTO TABLE t2 columns terminated by ',' (c0,c1,c2,c3) set (pk=c0, v1=c1, v2=c2, __op = c3)) with broker 'broker0'";
        SqlParser parser = new SqlParser(new SqlScanner(new StringReader(sql)));
        LoadStmt loadStmt = (LoadStmt) SqlParserUtils.getFirstStmt(parser);
        List<Expr> columnMappingList = Deencapsulation.getField(loadStmt.getDataDescriptions().get(0),
                "columnMappingList");

        // file groups
        List<BrokerFileGroup> fileGroups = Lists.newArrayList();
        List<String> files = Lists.newArrayList("/path/file1");
        List<String> columnNames = Lists.newArrayList("c0", "c1", "c2", "c3");
        DataDescription desc = new DataDescription("t2", null, files, columnNames,
                null, null, "CSV", Lists.newArrayList(),
                false, columnMappingList, null);
        Deencapsulation.invoke(desc, "analyzeColumns");
        BrokerFileGroup brokerFileGroup = new BrokerFileGroup(desc);
        Deencapsulation.setField(brokerFileGroup, "columnSeparator", ",");
        Deencapsulation.setField(brokerFileGroup, "rowDelimiter", "\n");
        Deencapsulation.setField(brokerFileGroup, "fileFormat", "CSV");
        fileGroups.add(brokerFileGroup);

        // file status
        List<List<TBrokerFileStatus>> fileStatusesList = Lists.newArrayList();
        List<TBrokerFileStatus> fileStatusList = Lists.newArrayList();
        fileStatusList.add(new TBrokerFileStatus("/path/file1", false, 128000000, true));
        fileStatusesList.add(fileStatusList);

        // plan
        Config.vectorized_load_enable = true;
        LoadingTaskPlanner planner = new LoadingTaskPlanner(jobId, txnId, db.getId(), table, brokerDesc, fileGroups,
                false, TimeUtils.DEFAULT_TIME_ZONE, 3600);
        planner.plan(loadId, fileStatusesList, 1);

        // 2. check scan node column expr
        FileScanNode scanNode = (FileScanNode) planner.getScanNodes().get(0);
        List<TScanRangeLocations> locationsList = scanNode.getScanRangeLocations(0);
        Assert.assertEquals(1, locationsList.size());
        TScanRangeLocations location = locationsList.get(0);
        TBrokerScanRangeParams params = location.scan_range.broker_scan_range.params;
        Map<Integer, TExpr> exprOfDestSlot = params.expr_of_dest_slot;

        // get last slot: id == 3
        TExpr opExpr = exprOfDestSlot.get(3);
        Assert.assertEquals(2, opExpr.nodes.size());
        Assert.assertEquals(TExprNodeType.CAST_EXPR, opExpr.nodes.get(0).node_type);
        Assert.assertEquals(TExprNodeType.SLOT_REF, opExpr.nodes.get(1).node_type);
    }

    @Test
    public void testLoadWithOpAutoMapping(@Mocked Catalog catalog, @Mocked SystemInfoService systemInfoService,
                                          @Injectable Database db, @Injectable OlapTable table) throws Exception {
        // table schema
        List<Column> columns = Lists.newArrayList();
        columns.add(new Column("pk", Type.BIGINT, true, null, false, "123", ""));
        columns.add(new Column("v1", Type.INT, false, null, false, "231", ""));
        columns.add(new Column("v2", ScalarType.createVarchar(50), false, null, true, "asdf", ""));

        Function f1 = new Function(new FunctionName("casttobigint"), new Type[] {Type.VARCHAR},
                Type.BIGINT, true);
        f1.setIsVectorized(true);
        Function f2 = new Function(new FunctionName("casttoint"), new Type[] {Type.VARCHAR},
                Type.INT, true);
        f2.setIsVectorized(true);
        Function f3 = new Function(new FunctionName("casttotinyint"), new Type[] {Type.VARCHAR},
                Type.TINYINT, true);
        f3.setIsVectorized(true);

        new Expectations() {
            {
                Catalog.getCurrentSystemInfo();
                result = systemInfoService;
                systemInfoService.getIdToBackend();
                result = idToBackend;
                table.getKeysType();
                result = KeysType.PRIMARY_KEYS;
                table.getBaseSchema();
                result = columns;
                table.getFullSchema();
                result = columns;
                table.getPartitions();
                minTimes = 0;
                result = Arrays.asList(partition);
                table.getColumn("pk");
                result = columns.get(0);
                table.getColumn("v1");
                result = columns.get(1);
                table.getColumn("v2");
                result = columns.get(2);
                table.getColumn(Load.LOAD_OP_COLUMN);
                result = null;
                returns(f1, f2, f3);
            }
        };

        // column mappings
        String sql =
                "LOAD LABEL label0 (DATA INFILE('/path/file1') INTO TABLE t2 columns terminated by ',' (pk,v1,v2,__op)) with broker 'broker0'";
        SqlParser parser = new SqlParser(new SqlScanner(new StringReader(sql)));
        LoadStmt loadStmt = (LoadStmt) SqlParserUtils.getFirstStmt(parser);
        List<Expr> columnMappingList = Deencapsulation.getField(loadStmt.getDataDescriptions().get(0),
                "columnMappingList");

        // file groups
        List<BrokerFileGroup> fileGroups = Lists.newArrayList();
        List<String> files = Lists.newArrayList("/path/file1");
        List<String> columnNames = Lists.newArrayList("pk", "v1", "v2", "__op");
        DataDescription desc = new DataDescription("t2", null, files, columnNames,
                null, null, "CSV", Lists.newArrayList(),
                false, columnMappingList, null);
        Deencapsulation.invoke(desc, "analyzeColumns");
        BrokerFileGroup brokerFileGroup = new BrokerFileGroup(desc);
        Deencapsulation.setField(brokerFileGroup, "columnSeparator", ",");
        Deencapsulation.setField(brokerFileGroup, "rowDelimiter", "\n");
        Deencapsulation.setField(brokerFileGroup, "fileFormat", "CSV");
        fileGroups.add(brokerFileGroup);

        // file status
        List<List<TBrokerFileStatus>> fileStatusesList = Lists.newArrayList();
        List<TBrokerFileStatus> fileStatusList = Lists.newArrayList();
        fileStatusList.add(new TBrokerFileStatus("/path/file1", false, 128000000, true));
        fileStatusesList.add(fileStatusList);

        // plan
        Config.vectorized_load_enable = true;
        LoadingTaskPlanner planner = new LoadingTaskPlanner(jobId, txnId, db.getId(), table, brokerDesc, fileGroups,
                false, TimeUtils.DEFAULT_TIME_ZONE, 3600);
        planner.plan(loadId, fileStatusesList, 1);

        // 2. check scan node column expr
        FileScanNode scanNode = (FileScanNode) planner.getScanNodes().get(0);
        List<TScanRangeLocations> locationsList = scanNode.getScanRangeLocations(0);
        Assert.assertEquals(1, locationsList.size());
        TScanRangeLocations location = locationsList.get(0);
        TBrokerScanRangeParams params = location.scan_range.broker_scan_range.params;
        Map<Integer, TExpr> exprOfDestSlot = params.expr_of_dest_slot;

        TExpr opExpr = exprOfDestSlot.get(3);
        Assert.assertEquals(1, opExpr.nodes.size());
        Assert.assertEquals(TExprNodeType.SLOT_REF, opExpr.nodes.get(0).node_type);
    }
}
